#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2017 Pengutronix, Chris Fiege <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import ctypes
import string

"""
This module contains functions related to packing data into ctype arrays in
special ways as needed by the Microchip USB2642.
"""


def string_to_uint8_array(str, array_length, c_string=False, padding=0xFF, encoding="UTF-16"):
    """
    Converts a python-string into a ctypes.c_uint8 array of a given length.
    The str will be encoded with the given encoding before converting.

    If c_string is True the string will be terminated with 0x00.
    str will be padded with padding if the buffer is longer than the str.
    str will be cropped if the buffer is shorter.

    Arguments:
    str -- python string
    array_length -- length of the resulting array in bytes.
    c_string -- Switch to treat a str as c-string. String will be terminated with 0x00
    padding -- This value will be used to pad buffer to the given length.
    """

    # Preparing output array
    byte_buf = (ctypes.c_uint8 * array_length)()
    for i in range(array_length):
        byte_buf[i] = padding

    # Determine number of bytes to copy
    nbytes = str.encode(encoding)
    length = array_length - 1 if c_string else array_length
    count = min(len(nbytes), length)

    # Do the actual copy
    for i in range(count):
        byte_buf[i] = int(nbytes[i])

    # Make string a c-string
    if c_string:
        i += 1
        byte_buf[i] = 0x00

    return byte_buf


def string_to_microchip_unicode_uint8_array(text, array_length, constant=0x03):
    """
    Converts a String to a USB2642 UTF-16 string.

    The USB2642 requires the first two bytes of the string to be
    <length including first two bytes><0x03>.
    This function first creates a UTF-16 string and then replaces the byte-order
    mark with this information.

    Arguments:
    text -- Text to copy into the array
    constant -- The constant byte placed into the 2nd byte
    """

    byte_buf = string_to_uint8_array(text, array_length)
    byte_buf[0] = len(text) * 2 + 2
    byte_buf[1] = constant
    return byte_buf


def list_to_uint8_array(numbers, array_length):
    """
    Converts a list of numbers into a ctypes.c_uint8 array of a given length.

    If numbers is too short for array_length it will be padded with 0x00.
    If numbers is too long it will be cropped.

    Arguments:
    numbers -- iterable of numbers (int, bytes, float...)
    array_length -- length of the resulting array
    """

    byte_buf = (ctypes.c_uint8 * array_length)()

    count = min(len(numbers), array_length)
    for i in range(count):
        byte_buf[i] = int(numbers[i])
    return byte_buf


def to_pretty_hex(buffer):
    """Takes a byte-buffer and creates a pretty-looking hex-string from it"""

    if isinstance(buffer, ctypes.Structure):
        out = ctypes.create_string_buffer(ctypes.sizeof(buffer))
        ctypes.memmove(ctypes.addressof(out), ctypes.addressof(buffer), ctypes.sizeof(buffer))
        temp_buf = [ord(x) for x in out]
    elif isinstance(buffer[0], int):
        temp_buf = [x for x in buffer]
    else:
        temp_buf = [ord(x) for x in buffer]

    res = ""
    offs = 0
    while len(temp_buf) > 0:
        window = temp_buf[0:8]
        temp_buf = temp_buf[8:]
        res += "0x{:02X}\t{}  {}\n".format(
            offs,
            " ".join([f"{x:02X}" for x in window]),
            " ".join([chr(x) if chr(x) in string.printable.split(" ")[0] else "." for x in window]),
        )
        offs += 8
    return res



#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2023 Pengutronix, Jan Lübbe <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA


def bitslice(val, upper, lower):
    "extract a contiguous slice of bits from a larger value"
    size = upper - lower + 1
    mask = (1 << size) - 1
    return (val >> lower) & mask


def decoded_to_text(decoded):
    "convert the output of the decode() function to human-readable text"
    text = []
    text.append(f"{decoded['reg']} Register Value: {decoded['raw']}")
    for f in decoded["fields"]:
        if "name" in f:
            text.append(f"  {f['field']}: {f['name']}")
        else:
            text.append(f"  {f['field']}")
        raw = f["raw"]
        text.append(f"    raw: 0b{format(raw[0], '0%db' % raw[1])} == 0x{raw[0]:0x} == {int(raw[0])}")
        if "enum" in f:
            text.append(f"    enum: {f['enum']} {f.get('unit', '')}".rstrip())
        if "bits" in f:
            text.append(f"    bits: {', '.join(f['bits'])}")
        if "decoded" in f:
            text.append(f"    decoded: {f['decoded']}")
        if "value" in f:
            text.append(f"    value: {f['value']} {f.get('unit', '')}".rstrip())
    return text


class RegisterDecoder:
    "decode a register based on a mostly declarative description of the contents."

    FIELDS = {}

    def __init__(self, raw_hex):
        self.raw_hex = raw_hex
        self.raw = int(raw_hex, 16)

    def _get_slice(self, field):
        field = self.FIELDS[field]
        return field["slice"]

    def _get_details(self, field):
        field = self.FIELDS[field]
        return field

    def _get_value(self, field):
        upper, lower = self._get_slice(field)
        width = upper - lower + 1
        value = bitslice(self.raw, upper, lower)
        return (value, width)

    def __getattr__(self, field):
        v, _ = self._get_value(field)
        details = self._get_details(field)
        if "convert" in details:
            v = details["convert"](v)
        return v

    def decode_field(self, field):
        "decode a single field into a key-value format"
        result = {}

        value, width = self._get_value(field)
        details = self._get_details(field)
        result["field"] = field
        if "name" in details:
            result["name"] = details["name"]
        result["raw"] = (value, width)
        if "enum" in details:
            try:
                result["enum"] = details["enum"][value]
            except (IndexError, KeyError):
                result["enum"] = None
        if "bits" in details:
            bits = [details["bits"][x] for x in range(width) if (value & (1 << x))]
            result["bits"] = bits
        if "convert" in details:
            result["value"] = str(details["convert"](value))
        if "unit" in details:
            result["unit"] = details["unit"]
        if "decode" in details:
            decoded = details["decode"](self)
            result.update(decoded)

        return result

    def get_computed(self):
        "compute additional values which need multiple fields as input"
        return {}

    def decode(self):
        "decode all fields into a format suitable for JSON encoding"
        result = {"reg": self.__class__.__name__, "raw": self.raw_hex, "fields": []}

        for field, _ in sorted(self.FIELDS.items(), key=lambda x: x[1]["slice"], reverse=True):
            result["fields"].append(self.decode_field(field))
        result["computed"] = self.get_computed()

        return result

    def get_text_report(self):
        "decode all fields and format as human-readable text"
        decoded = self.decode()
        return decoded_to_text(decoded)


class CSD_Common(RegisterDecoder):
    TIME_VALUE_ENUM = ["reserved", 1.0, 1.2, 1.3, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0, 7.0, 8.0]

    def decode_TAAC(self):
        v = self._get_value("TAAC")
        TIME_UNIT_ENUM = ["1ns", "10ns", "100ns", "1us", "10us", "100us", "1ms", "10ms"]
        TIME_SCALE_ENUM = [1, 10, 100, 1_000, 10_000, 100_000, 1_000_000, 10_000_000]
        unit = TIME_UNIT_ENUM[bitslice(v[0], 2, 0)]
        scale = TIME_SCALE_ENUM[bitslice(v[0], 2, 0)]
        value = self.TIME_VALUE_ENUM[bitslice(v[0], 6, 3)]

        scaled_value = value * scale if isinstance(value, float) else None

        return {
            "decoded": (value, unit),
            "value": scaled_value,
            "unit": "ns",
        }

    def decode_TRAN_SPEED(self):
        v = self._get_value("TRAN_SPEED")
        RATE_UNIT_ENUM = ["100 Kbit/s", "1Mbit/s", "10Mbit/s", "100Mbit/s", "reserved"]
        RATE_SCALE_ENUM = [100_000, 1_000_000, 10_000_000, 10_000_000, "reserved"]
        unit = RATE_UNIT_ENUM[bitslice(v[0], 2, 0)]
        scale = RATE_SCALE_ENUM[bitslice(v[0], 2, 0)]
        value = self.TIME_VALUE_ENUM[bitslice(v[0], 6, 3)]

        scaled_value = None

        if isinstance(value, float) and isinstance(scale, int):
            scaled_value = value * scale

        return {
            "decoded": (value, unit),
            "value": scaled_value,
            "unit": "bit/s",
        }

    FIELDS = {
        "CSD_STRUCTURE": {
            "slice": (127, 126),
            "name": "CSD structure",
            "enum": ["1.0", "2.0"],
        },
        "TAAC": {
            "slice": (119, 112),
            "name": "data read access-time-1",
            "decode": decode_TAAC,
        },
        "NSAC": {
            "slice": (111, 104),
            "name": "data read access-time-2",
            "convert": lambda v: v * 100,
            "unit": "CLK cycles",
        },
        "TRAN_SPEED": {
            "slice": (103, 96),
            "name": "max. data transfer rate",
            "decode": decode_TRAN_SPEED,
        },
        "CCC": {"slice": (95, 84), "name": "card command classes", "bits": [str(x) for x in range(12)]},
        "READ_BL_LEN": {
            "slice": (83, 80),
            "name": "max. read data block length",
            "convert": lambda v: 2**v,
            "unit": "bytes",
        },
        "READ_BL_PARTIAL": {
            "slice": (79, 79),
            "name": "partial blocks for read allowed",
            "convert": bool,
        },
        "WRITE_BLK_MISALIGN": {
            "slice": (78, 78),
            "name": "write block misalignment allowed",
            "convert": bool,
        },
        "READ_BLK_MISALIGN": {
            "slice": (77, 77),
            "name": "read block misalignment allowed",
            "convert": bool,
        },
        "DSR_IMP": {
            "slice": (76, 76),
            "name": "driver stage register implemented",
            "convert": bool,
        },
        "ERASE_BLK_EN": {
            "slice": (46, 46),
            "name": "erase single block enable",
            "convert": bool,
        },
        "SECTOR_SIZE": {
            "slice": (45, 39),
            "name": "erase sector size",
            "convert": lambda v: v + 1,
            "unit": "write blocks",
        },
        "WP_GRP_SIZE": {
            "slice": (38, 32),
            "name": "write protect group size",
            "convert": lambda v: v + 1,
            "unit": "erase sectors",
        },
        "WP_GRP_ENABLE": {
            "slice": (31, 31),
            "name": "write protect group enable",
            "convert": bool,
        },
        "R2W_FACTOR": {
            "slice": (28, 26),
            "name": "write speed factor",
            "convert": lambda v: 2**v,
            "unit": "multiples of read access time",
        },
        "WRITE_BL_LEN": {
            "slice": (25, 22),
            "name": "max. write data block length",
            "convert": lambda v: 2**v,
            "unit": "bytes",
        },
        "WRITE_BL_PARTIAL": {
            "slice": (21, 21),
            "name": "partial blocks for write allowed",
            "convert": bool,
        },
        "FILE_FORMAT_GRP": {
            "slice": (15, 15),
            "name": "file format group",
        },
        "COPY": {
            "slice": (14, 14),
            "name": "copy flag",
            "convert": bool,
        },
        "PERM_WRITE_PROTECT": {
            "slice": (13, 13),
            "name": "permanent write protection",
            "convert": bool,
        },
        "TMP_WRITE_PROTECT": {
            "slice": (12, 12),
            "name": "temporary write protection",
            "convert": bool,
        },
        "FILE_FORMAT": {
            "slice": (11, 9),
            "name": "file format",
            "enum": [
                "Hard disk-like file system with partition table",
                "DOS FAT (floppy-like) with boot sector only (no partition table)",
                "Universal File Format",
                "Others/Unknown",
            ],
        },
        "CRC": {
            "slice": (7, 1),
        },
    }

    def get_computed(self):
        result = {}

        sector_size = self.SECTOR_SIZE * self.WRITE_BL_LEN
        wp_grp_size = self.WP_GRP_SIZE * sector_size

        result["sector_size_bytes"] = {
            "value": sector_size,
            "unit": "bytes",
            "name": "sector size",
        }
        result["wp_grp_size_bytes"] = {
            "value": wp_grp_size,
            "unit": "bytes",
            "name": "write protect group size",
        }

        return result


class CSD_10(CSD_Common):
    VDD_MIN_CURR_MA_ENUM = [0.5, 1, 5, 10, 25, 35, 60, 100]
    VDD_MAX_CURR_MA_ENUM = [1, 5, 10, 25, 35, 45, 80, 200]

    FIELDS = CSD_Common.FIELDS.copy()
    FIELDS["C_SIZE"] = {
        "slice": (73, 62),
        "name": "device size",
        "convert": lambda v: (v + 1),
    }
    FIELDS["VDD_R_CURR_MIN"] = {
        "slice": (61, 59),
        "name": "max. read current @VDD min",
        "enum": VDD_MIN_CURR_MA_ENUM,
        "unit": "mA",
    }
    FIELDS["VDD_R_CURR_MAX"] = {
        "slice": (58, 56),
        "name": "max. read current @VDD max",
        "enum": VDD_MAX_CURR_MA_ENUM,
        "unit": "mA",
    }
    FIELDS["VDD_W_CURR_MIN"] = {
        "slice": (55, 53),
        "name": "max. write current @VDD min",
        "enum": VDD_MIN_CURR_MA_ENUM,
        "unit": "mA",
    }
    FIELDS["VDD_W_CURR_MAX"] = {
        "slice": (52, 50),
        "name": "max. write current @VDD max",
        "enum": VDD_MAX_CURR_MA_ENUM,
        "unit": "mA",
    }
    FIELDS["C_SIZE_MULT"] = {"slice": (49, 47), "name": "device size multiplier", "convert": lambda v: 2 ** (v + 2)}

    def get_computed(self):
        result = super().get_computed()

        device_size = self.C_SIZE * self.C_SIZE_MULT * self.READ_BL_LEN

        result["device_size_bytes"] = {
            "value": device_size,
            "unit": "bytes",
            "name": "device size",
        }

        return result


class CSD_20(CSD_Common):
    FIELDS = CSD_Common.FIELDS.copy()
    FIELDS["C_SIZE"] = {
        "slice": (69, 48),
        "name": "device size",
        "convert": lambda v: (v + 1) * 512 * 1024,
        "unit": "bytes",
    }


def decode_csd(raw_hex):
    raw = int(raw_hex, 16)
    val = bitslice(raw, 127, 126)
    if val == 0:
        return CSD_10(raw_hex)
    elif val == 1:
        return CSD_20(raw_hex)
    else:
        raise ValueError(f"unknown CSD version {val}")


class CID(RegisterDecoder):
    @staticmethod
    def convert_printable(value, size):
        # The spec defines some fields as ASCII strings, but in practice, cards
        # contain many other values as well. Clean up non-printable characters,
        # newlines, etc.
        result = ""
        for i in reversed(range(size)):
            c = chr((value >> 8 * i) & 0x7F)
            result += c if c.isprintable() else "."
        return result

    FIELDS = {
        "MID": {
            "slice": (127, 120),
            "name": "Manufacturer ID",
            "enum": {
                0x02: "SanDisk",
                0x03: "SanDisk SD",
                0x1B: "Samsung",
                0x74: "Transcend",
                0x9F: "Kingston SD",
            },
        },
        "OID": {
            "slice": (119, 104),
            "name": "OEM/Application ID",
            "convert": lambda v: CID.convert_printable(v, 2),
        },
        "PNM": {
            "slice": (103, 64),
            "name": "Product name",
            "convert": lambda v: CID.convert_printable(v, 5),
        },
        "PRV": {
            "slice": (63, 56),
            "name": "Product revision",
            "convert": lambda v: f"{int(v >> 4 & 15)}.{int(v & 15)}",
        },
        "PSN": {
            "slice": (55, 24),
            "name": "Product serial number",
            "convert": int,
        },
        "RESERVED": {
            "slice": (23, 20),
        },
        # "MDT": {  # combined layout for reference
        #    "slice": (19, 8),
        #    "name": "Manufacturing date",
        # },
        "MDT_Y": {
            "slice": (19, 12),
            "name": "Manufacturing date (year)",
            "convert": lambda v: v + 2000,
        },
        "MDT_M": {
            "slice": (11, 8),
            "name": "Manufacturing date (month)",
            "convert": lambda v: v + 1,
        },
        "CRC": {
            "slice": (7, 1),
            "name": "CRC7 checksum",
        },
        "NU1": {
            "slice": (0, 0),
            "name": "not used, always 1",
        },
    }


class SCR(RegisterDecoder):
    FIELDS = {
        "SCR_STRUCTURE": {
            "slice": (63, 60),
            "name": "SCR Structure",
            "enum": ["1.0"],
        },
        "SD_SPEC": {
            "slice": (59, 56),
            "name": "SD Memory Card - Spec. Version",
            "enum": ["1.0 or 1.01", "1.10", "2.00 or 3.0X"],
        },
        "DATA_STAT_AFTER_ERASE": {
            "slice": (55, 55),
            "name": "data status after erase",
            "convert": bool,
        },
        "SD_SECURITY": {
            "slice": (54, 52),
            "name": "CPRM Security Support",
            "enum": [
                "No Security",
                "Not Used",
                "SDSC Card (Security Version 1.01)",
                "SDHC Card (Security Version 2.00)",
                "SDXC Card (Security Version 3.xx)",
            ],
        },
        "SD_BUS_WIDTHS": {
            "slice": (51, 48),
            "name": "DAT Bus widths supported",
            "bits": ["1 bit", None, "4 bit", None],
        },
        "SD_SPEC3": {
            "slice": (47, 47),
            "name": "Spec. Version 3.00 or higher",
        },
        "EX_SECURITY": {
            "slice": (46, 43),
            "name": "Extended Security Support",
        },
        "RESERVED": {
            "slice": (42, 34),
        },
        "CMD_SUPPORT": {
            "slice": (33, 32),
            "name": "Command Support bits",
            "bits": ["Speed Class Control (CMD20)", "Set Block Count (CMD23)"],
        },
        "RESERVED_MFG": {
            "slice": (31, 0),
        },
    }


if __name__ == "__main__":
    import argparse
    import json

    parser = argparse.ArgumentParser()
    parser.add_argument("--json", action="store_true")
    parser.add_argument("input", help="input in JSON format", type=str)

    args = parser.parse_args()

    with open(args.input) as fd:
        data = json.loads(fd.read())

    if args.json:
        res = {}
        res["scr"] = SCR(data["scr"]["raw"]).decode()
        res["cid"] = CID(data["cid"]["raw"]).decode()
        res["csd"] = decode_csd(data["csd"]["raw"]).decode()
        print(json.dumps(res, indent=2))
    else:
        res = []
        res += decode_csd(data["csd"]["raw"]).get_text_report()
        res += SCR(data["scr"]["raw"]).get_text_report()
        res += CID(data["cid"]["raw"]).get_text_report()
        print("\n".join(res))



#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2017 Pengutronix, Chris Fiege <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

from .usb2642 import Usb2642


class I2cGpio:
    # Registers inside the supported GPIO expanders
    _register_inputPort = 0x00
    _register_outputPort = 0x01
    _register_polarity = 0x02
    _register_configuration = 0x03

    # Values for the configuration register
    _direction_output = 0
    _direction_input = 1

    def __init__(self, sg):
        """
        Arguments:
        sg -- /dev/sg* to use.
        """
        self._usb = Usb2642(sg)

    def get_usb(self):
        return self._usb

    def _write_register(self, register, value):
        """
        Writes a register on the GPIO-expander with a given value.
        """
        self._usb.write_to(self._I2cAddr, [register, value])

    def _read_register(self, addr):
        """
        Returns a register of the GPIO-expander.
        """
        return self._usb.write_read_to(self._I2cAddr, [addr], 1)[0]

    def set_pin_to_output(self, pins):
        """
        Sets the corresponding pins as outputs.

        Arguments:
        pins -- Combination of I2cGpio.gpio_*
        """
        direction = self._read_register(self._register_configuration)
        direction = (direction & ~pins) & 0xFF
        self._write_register(self._register_configuration, direction)

    def set_pin_to_input(self, pins):
        """
        Sets the corresponding pins as inputs.

        Arguments:
        pins -- Combination of I2cGpio.gpio_*
        """
        direction = self._read_register(self._register_configuration)
        direction = direction | pins
        self._write_register(self._register_configuration, direction)

    def get_gpio_config(self):
        """
        Returns the state of the configuration register.
        """
        return self._read_register(self._register_configuration)

    def get_input_values(self):
        """
        Reads the value currently present on the port from the input register.
        """
        return self._read_register(self._register_inputPort)

    def output_values(self, values: int, bitmask: int = 0xFF):
        """
        Writes the given values to the GPIO-expander.
        Pins configured as Inputs are not affected by this.

        Arguments:
        values -- Combination of I2cGpio.gpio_*
        bitmask -- Only update bits in the register that are '1' in the bitmask
        """

        if bitmask == 0xFF:
            # trivial case: Let's just write the value
            self._write_register(self._register_outputPort, values)
        else:
            # complex case: Let's do a read-modify-write
            val = self._read_register(self._register_outputPort)
            val = (val & ~bitmask) & 0xFF  # reset masked bits
            val = val | (values & bitmask)  # set bits set in values and bitmask
            self._write_register(self._register_outputPort, val)


class Pca9536(I2cGpio):
    """
    Interface to control a Pca9536 that is connected to the auxiliary-I2C of a
    Microchip USB2642.
    """

    # The PCA9536 I2C slave Address in 7-Bit Format
    _I2cAddr = 0b100_0001

    gpio_0 = 0x01
    gpio_1 = 0x02
    gpio_2 = 0x04
    gpio_3 = 0x08


class Tca6408(I2cGpio):
    """
    Interface to control a TCA6408 that is connected to the auxiliary-I2C of a
    Microchip USB2642.
    """

    # The TCA6408 I2C slave Address in 7-Bit Format
    _I2cAddr = 0b010_0000

    gpio_0 = 0x01
    gpio_1 = 0x02
    gpio_2 = 0x04
    gpio_3 = 0x08
    gpio_4 = 0x10
    gpio_5 = 0x20
    gpio_6 = 0x40
    gpio_7 = 0x80



#! /usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2017 Pengutronix, Chris Fiege <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import argparse
import errno
import json
import sys

from .mqtthelper import Config, publish_info
from .sd_regs import decoded_to_text
from .usbsdmux import NotInHostModeException, UnknownUsbSdMuxRevisionException, autoselect_driver


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument("sg", metavar="SG", help="/dev/sg* to use")

    parser.add_argument("--config", help="Set config file location", default=None)

    format_parser = parser.add_mutually_exclusive_group()
    format_parser.add_argument("--json", help="Format output as json. Useful for scripting.", action="store_true")

    subparsers = parser.add_subparsers(help="Supply one of the following commands to interact with the device")
    subparsers.required = True
    subparsers.dest = "mode"

    subparsers.add_parser("get", help="Read the current state of the USB-SD-Mux")
    subparsers.add_parser("dut", help="Switch to the DUT")
    subparsers.add_parser("client", help="Switch to the DUT")
    subparsers.add_parser("host", help="Switch to the host")
    subparsers.add_parser("off", help="Disconnect from host and DUT")

    parser_gpio = subparsers.add_parser("gpio", help="Manipulate a GPIO (open drain output only)")
    parser_gpio.add_argument("gpio", help="The GPIO to change", choices=[0, 1], type=int)
    parser_gpio.add_argument("action", help="What to do with the GPIO", choices=["low", "0", "high", "1", "get"])

    subparsers.add_parser("info", help="Show information about the SD card")

    args = parser.parse_args()

    config = Config(args.config)

    try:
        ctl = autoselect_driver(args.sg)
    except UnknownUsbSdMuxRevisionException as e:
        error_msg = str(e) + "\n" + f"Does {args.sg} really point to a USB-SD-Mux?"
        if args.json:
            print(json.dumps({"error-message": error_msg}))
        else:
            print(error_msg, file=sys.stderr)
        sys.exit(1)
    mode = args.mode

    error_msg = None
    try:
        if mode == "off":
            ctl.mode_disconnect()
            if args.json:
                print(json.dumps({}))

        elif mode in ("dut", "client"):
            publish_info(ctl, config, args.sg, "client")

            ctl.mode_DUT()
            if args.json:
                print(json.dumps({}))

        elif mode == "host":
            ctl.mode_host()
            if args.json:
                print(json.dumps({}))

            publish_info(ctl, config, args.sg, "host")

        elif mode == "get":
            if args.json:
                print(json.dumps({"switch-state": ctl.get_mode()}))
            else:
                print(ctl.get_mode())

        elif mode == "gpio":
            if args.action == "get":
                if args.json:
                    print(json.dumps({"gpio-state": {"gpio": args.gpio, "state:": ctl.gpio_get(args.gpio)}}))
                else:
                    print(ctl.gpio_get(args.gpio))
            elif args.action in ["0", "low"]:
                ctl.gpio_set_low(args.gpio)
                if args.json:
                    print(json.dumps({}))
            elif args.action in ["1", "high"]:
                ctl.gpio_set_high(args.gpio)
                if args.json:
                    print(json.dumps({}))

        elif mode == "info":
            info = ctl.get_card_info()
            if args.json:
                print(json.dumps(info, indent=2))
            else:
                print("\n".join(decoded_to_text(info["scr"])))
                print("\n".join(decoded_to_text(info["cid"])))
                print("\n".join(decoded_to_text(info["csd"])))

    except FileNotFoundError as fnfe:
        error_msg = str(fnfe)
    except PermissionError as perr:
        error_msg = str(perr)
    except OSError as ose:
        if ose.errno == errno.ENOTTY:
            # ENOTTY is raised when an error occurred when calling an ioctl
            error_msg = ose + "\n" + f"Does '{args.sg}' really point to a USB-SD-Mux?"
        else:
            raise ose
    except NotInHostModeException:
        error_msg = "Card information is only available in host mode."
    except NotImplementedError:
        error_msg = "This USB-SD-Mux does not support GPIOs."

    if error_msg:
        if args.json:
            print(json.dumps({"error-message": error_msg}))
        else:
            print(error_msg, file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



import configparser
import json
import os
import sys


class Config:
    """
    Reads the configuration file by default at /etc/usbsdmux.config
    """

    def __init__(self, configfile):
        if configfile is not None:
            if not os.path.isfile(configfile):
                raise FileNotFoundError("Config file {configfile} not found")
        else:
            configfile = "/etc/usbsdmux.config"

        config = configparser.ConfigParser()
        config.read(configfile)

        if "mqtt" not in config or "send" not in config:
            self.mqtt_enabled = False
            return
        else:
            self.mqtt_enabled = True

        mqtt_section = config["mqtt"]

        for argument in ("server", "port", "topic"):
            if argument not in mqtt_section:
                raise ValueError(f"Config value mqtt/{argument} not found. Please check {configfile}")

        self.mqtt_server = mqtt_section["server"]
        self.mqtt_port = int(mqtt_section["port"])
        self.mqtt_topic = mqtt_section["topic"]

        if "username" in mqtt_section and "password" in mqtt_section:
            self.mqtt_auth = {"username": mqtt_section["username"], "password": mqtt_section["password"]}
        else:
            self.mqtt_auth = None

        send_section = config["send"]

        self.send_on_host = send_section.get("host", False)
        self.send_on_dut = send_section.get("dut", False)


def _read_file(filename):
    try:
        with open(filename) as f:
            return f.read()
    except FileNotFoundError:
        return None


def _read_int(filename, base=10):
    try:
        return int(_read_file(filename).strip(), base)
    except TypeError:
        return None


def _gather_data(ctl, sg, mode):
    import socket

    import pkg_resources

    base_sg = os.path.realpath(sg)
    sg_name = os.path.basename(base_sg)

    # only file in this directory is a hard link pointing to the block device
    sd_name = os.listdir(f"/sys/class/scsi_generic/{sg_name}/device/block/")[0]

    # using that name we can obtain further information
    stat_data = [int(part) for part in _read_file(f"/sys/class/block/{sd_name}/stat").split()]

    # https://www.kernel.org/doc/Documentation/ABI/testing/sysfs-block
    stat_names = (
        "reads_completed_successfully",
        "reads_merged",
        "sectors_read",
        "time_spent_reading",
        "writes_completed",
        "writes_merged",
        "sectors_written",
        "time_spent_writing",
        "IOs_currently_in_progress",
        "time_spent_doing_IOs",
        "weighted_time_spent_doing_IOs",
        "discards_completed",
        "discards_merged",
        "sectors_discarded",
        "time_spent_discarding",
        "flush_requests_completed",
        "time_spent_flushing",
    )

    stat = dict(zip(stat_names, stat_data))

    usb_path = os.path.realpath(f"/sys/class/scsi_generic/{sg_name}")

    max_depth = 10
    while not os.path.isfile(os.path.join(usb_path, "serial")) and max_depth > 0:
        usb_path = os.path.dirname(usb_path)
        max_depth -= 1

    card_info = ctl.get_card_info() if ctl.get_mode() == "host" or mode == "dut" else None

    data = {
        "command": " ".join(sys.argv),
        "mode": mode,
        "sg": sg_name,
        "sd": sd_name,
        "usb": usb_path,
        "username": os.getlogin(),
        "hostname": socket.gethostname(),
        "labgrid-place": os.environ.get("LG_PLACE"),
        "model": type(ctl).__name__,
        "serial": _read_file(os.path.join(usb_path, "serial")).strip(),
        "version": pkg_resources.get_distribution("usbsdmux").version,
        "diskseq": _read_int(f"/sys/class/block/{sd_name}/diskseq"),
        "size": _read_int(f"/sys/class/block/{sd_name}/size"),
        "ioerr_cnt": _read_int(f"/sys/class/scsi_generic/{sg_name}/device/ioerr_cnt", 16),
        "stat": stat,
        "card_info": card_info,
    }

    return data


def publish_info(ctl, config, sg, mode):
    """
    Publish info to mqtt server, if mqtt is enabled.
    This requires installing paho-mqtt.
    """

    if not config.mqtt_enabled:
        return

    if (mode == "client" and not config.send_on_dut) or (mode == "host" and not config.send_on_host):
        return

    try:
        import paho.mqtt.publish as mqtt
    except ImportError:
        print(
            "Sending data to an mqtt server requires paho-mqtt",
            "Please install it, e.g. by installing usbsdmux via:",
            "",
            '    python3 -m pip install "usbsdmux[mqtt]"',
            sep="\n",
            file=sys.stderr,
        )
        exit(1)

    data = _gather_data(ctl, sg, mode)
    mqtt.single(
        config.mqtt_topic,
        payload=json.dumps(data),
        hostname=config.mqtt_server,
        port=config.mqtt_port,
        auth=config.mqtt_auth,
    )



#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2017 Pengutronix, Chris Fiege <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import os
import time

from . import sd_regs
from .i2c_gpio import Pca9536, Tca6408


class UnknownUsbSdMuxRevisionException(Exception):
    pass


class NotInHostModeException(Exception):
    pass


def autoselect_driver(sg):
    """
    Create a new UsbSdMux with the correct driver for the device at /dev/<sg>

    Arguments:
    sg -- /dev/sg* to use
    """

    base_sg = os.path.realpath(sg)
    sg_name = os.path.basename(base_sg)
    model_filename = f"/sys/class/scsi_generic/{sg_name}/device/model"
    try:
        with open(model_filename) as fh:
            model = fh.read().strip()
        if model == "sdmux HS-SD/MMC":
            return UsbSdMuxClassic(sg)
        elif model == "sdFST HS-SD/MMC":
            return UsbSdMuxFast(sg)
        else:
            raise UnknownUsbSdMuxRevisionException(
                f"Could not determine type of USB-SD-Mux. Found unknown SCSI model '{model}'."
            )
    except FileNotFoundError as e:
        raise UnknownUsbSdMuxRevisionException(
            f"Could not determine type of USB-SD-Mux. Does {model_filename} exist?"
        ) from e


class UsbSdMux:
    """
    Class to provide an interface for the multiplexer on an usb-sd-mux.
    """

    def __init__(self, sg):
        """
        Create a new UsbSdMux.

        Arguments:
        sg -- /dev/sg* to use
        """
        raise NotImplementedError()

    def get_mode(self):
        """
        Returns currently selected mode as string
        """
        raise NotImplementedError()

    def mode_disconnect(self, wait=True):
        """
        Will disconnect the Micro-SD Card from both host and DUT.

        Arguments:
        wait -- Command will block for some time until the voltage-supply of
        the sd-card is known to be close to zero
        """
        raise NotImplementedError()

    def mode_DUT(self, wait=True):
        """
        Switches the MicroSD-Card to the DUT.

        This Command will issue a disconnect first to make sure the SD-card
        has been properly disconnected from both sides and its supply was off.
        """
        raise NotImplementedError()

    def mode_host(self, wait=True):
        """
        Switches the MicroSD-Card to the Host.

        This Command will issue a disconnect first to make sure the SD-card
        has been properly disconnected from both sides and its supply was off.
        """
        raise NotImplementedError()

    def gpio_get(self, gpio):
        """
        Reads the value of gpio and returns "high" or "low"
        """
        raise NotImplementedError()

    def gpio_set_high(self, gpio):
        """
        Sets a gpio high.
        """
        raise NotImplementedError()

    def gpio_set_low(self, gpio):
        """
        Sets a gpio low.
        """
        raise NotImplementedError()

    def get_card_info(self):
        if self.get_mode() != "host":
            raise NotInHostModeException()

        result = {}

        scr = self._usb.read_scr()
        result["scr"] = sd_regs.SCR(scr.hex()).decode()

        cid = self._usb.read_cid()
        result["cid"] = sd_regs.CID(cid.hex()).decode()

        csd = self._usb.read_csd()
        result["csd"] = sd_regs.decode_csd(csd.hex()).decode()

        return result


class UsbSdMuxClassic(UsbSdMux):
    _DAT_enable = 0x00
    _DAT_disable = Pca9536.gpio_0

    _PWR_enable = 0x00
    _PWR_disable = Pca9536.gpio_1

    _select_DUT = Pca9536.gpio_2
    _select_HOST = 0x00

    _card_inserted = 0x00
    _card_removed = Pca9536.gpio_3

    def __init__(self, sg):
        self._pca = Pca9536(sg)
        self._usb = self._pca.get_usb()

    def get_mode(self):
        val = self._pca.get_input_values()

        # If the SD-Card is disabled, we do not need to check for the selected mode.
        # PWR_disable and DAT_disable are always switched at the same time.
        # Let's assume it is sufficient to check one of both.
        if val & self._PWR_disable:
            return "off"

        if val & self._select_DUT:
            return "dut"

        return "host"

    def mode_disconnect(self, wait=True):
        # Set the output registers to known values and activate them afterward
        self._pca.output_values(self._DAT_disable | self._PWR_disable | self._select_HOST | self._card_removed)
        self._pca.set_pin_to_output(Pca9536.gpio_0 | Pca9536.gpio_1 | Pca9536.gpio_2 | Pca9536.gpio_3)

        if wait:
            time.sleep(1)

    def mode_DUT(self, wait=True):
        self.mode_disconnect(wait)

        # switch selection to DUT first to prevent glitches on power and
        # data-lines
        self._pca.output_values(self._DAT_disable | self._PWR_disable | self._select_DUT | self._card_removed)

        # now connect data and power
        self._pca.output_values(self._DAT_enable | self._PWR_enable | self._select_DUT | self._card_removed)

    def mode_host(self, wait=True):
        self.mode_disconnect(wait)

        # The disconnect-command has already switched the card to the host.
        # Thus, we don't need to worry about glitches anymore.

        # now connect data and power
        self._pca.output_values(self._DAT_enable | self._PWR_enable | self._select_HOST | self._card_inserted)


class UsbSdMuxFast(UsbSdMux):
    _DAT_enable = 0x00
    _DAT_disable = Tca6408.gpio_2

    _PWR_enable = 0x00
    _PWR_disable = Tca6408.gpio_1

    _select_DUT = Tca6408.gpio_0
    _select_HOST = 0x00

    _card_inserted = 0x00
    _card_removed = Tca6408.gpio_3

    gpio0 = Tca6408.gpio_5
    gpio1 = Tca6408.gpio_4

    def __init__(self, sg):
        self._tca = Tca6408(sg)
        self._assure_default_state()
        self._usb = self._tca.get_usb()

    def _assure_default_state(self):
        # If the USB-SD-Mux has just been powered on, its default ("DUT") is defined by pull-resistors.
        # If we now do a "read-modify-write" without taking into account the external default, we will
        # lose this state.
        # So let's check if the GPIO-expander is in Power-On-Reset defaults.
        # If so, write the same state to the device - but with driven outputs.

        if self._tca.get_gpio_config() == 0xFF:
            # If all pins are still set to "input" we are in the default state.
            # Let's set an output value that matches this configuration and set the relevant pins to output.
            self._tca.output_values(self._DAT_enable | self._PWR_enable | self._select_DUT | self._card_removed)
            self._tca.set_pin_to_output(
                Tca6408.gpio_0 | Tca6408.gpio_1 | Tca6408.gpio_2 | Tca6408.gpio_3 | Tca6408.gpio_4 | Tca6408.gpio_5
            )

    def get_mode(self):
        val = self._tca.get_input_values()

        # If the SD-Card is disabled, we do not need to check for the selected mode.
        # PWR_disable and DAT_disable are always switched at the same time.
        # Let's assume it is sufficient to check one of both.
        if val & self._PWR_disable:
            return "off"

        if val & self._select_DUT:
            return "dut"

        return "host"

    def mode_disconnect(self, wait=True):
        self._tca.output_values(
            values=self._DAT_disable | self._PWR_disable | self._select_HOST | self._card_removed,
            bitmask=self._tca.gpio_0 | self._tca.gpio_1 | self._tca.gpio_2 | self._tca.gpio_3,
        )

        if wait:
            time.sleep(1)

    def mode_DUT(self, wait=True):
        self.mode_disconnect(wait)

        # switch selection to DUT first to prevent glitches on power and
        # data-lines
        self._tca.output_values(
            values=self._DAT_disable | self._PWR_disable | self._select_DUT | self._card_removed,
            bitmask=self._tca.gpio_0 | self._tca.gpio_1 | self._tca.gpio_2 | self._tca.gpio_3,
        )

        # now connect data and power
        self._tca.output_values(
            values=self._DAT_enable | self._PWR_enable | self._select_DUT | self._card_removed,
            bitmask=self._tca.gpio_0 | self._tca.gpio_1 | self._tca.gpio_2 | self._tca.gpio_3,
        )

    def mode_host(self, wait=True):
        self.mode_disconnect(wait)

        # The disconnect-command has already switched the card to the host.
        # Thus, we don't need to worry about glitches anymore.

        # now connect data and power
        self._tca.output_values(
            values=self._DAT_enable | self._PWR_enable | self._select_HOST | self._card_inserted,
            bitmask=self._tca.gpio_0 | self._tca.gpio_1 | self._tca.gpio_2 | self._tca.gpio_3,
        )

    @staticmethod
    def _map_gpio(gpio):
        if gpio == 0:
            return UsbSdMuxFast.gpio0
        elif gpio == 1:
            return UsbSdMuxFast.gpio1
        raise ValueError("Unknown GPIO")

    def gpio_get(self, gpio):
        """
        Reads the value of gpio and returns "high" or "low"
        """
        gpio = UsbSdMuxFast._map_gpio(gpio)
        val = self._tca.get_input_values()
        if val & gpio:
            return "low"
        return "high"

    def gpio_set_high(self, gpio):
        gpio = UsbSdMuxFast._map_gpio(gpio)
        self._tca.output_values(0x0, gpio)

    def gpio_set_low(self, gpio):
        gpio = UsbSdMuxFast._map_gpio(gpio)
        self._tca.output_values(gpio, gpio)



#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2017 Pengutronix, Chris Fiege <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import argparse
import ctypes
import time

from .ctypehelper import (
    list_to_uint8_array,
    string_to_microchip_unicode_uint8_array,
    string_to_uint8_array,
)
from .usb2642 import Usb2642

"""
This module provides the high-level interface needed to write the contents of
the configuration-EEPROM of a USB2642 using the USB2642.
"""


class VerificationFailedException(Exception):
    pass


class USB2642Eeprom:
    """
    Provides an interface to write the configuration EEPROM of a USB2642.
    """

    def __init__(self, sg, i2c_addr=0x50):
        """
        Create a new USB2642Eeprom Instance.

        Arguments:
        sg -- /dev/sg* to use
        i2c_addr -- 7-Bit Address of the EEPROM to use. Defaults to 0x50 for the
                    configuration-EEPROM. You probably do NOT want to override this.
        """
        self.i2c = Usb2642(sg)
        self.addr = i2c_addr

    class _EepromStruct(ctypes.Structure):
        """
        Struct that contains the Configuration of the Card Reader and USB Hub.
        """

        _pack_ = 1  # forces the struct to be packed tight and overrides the default
        # 4-byte aligned packing.
        _fields_ = [
            # Flash Media Controller Configuration
            ("USB_SER_NUM", ctypes.c_uint8 * (0x19 - 0x00 + 1)),  # 0x00 .. 0x19 USB Serial Number
            ("USB_VID", ctypes.c_uint16),  # 0x1A .. 0x1B USB Vendor ID
            ("USB_PID", ctypes.c_uint16),  # 0x1C .. 0x1D USB Product ID
            (
                "USB_LANG_ID",
                ctypes.c_uint8 * (0x21 - 0x1E + 1),
            ),  # 0x1E  .. 0x21 USB Language Identifier, see http://www.usb.org/developers/docs/USB_LANGIDs.pdf
            ("USB_MFR_STR", ctypes.c_uint8 * (0x5D - 0x22 + 1)),  # 0x22 .. 0x5D USB Manufacturer String (Unicode)
            ("USB_PRD_STR", ctypes.c_uint8 * (0x99 - 0x5E + 1)),  # 0x5E .. 0x99 USB Product String (Unicode)
            (
                "USB_BM_ATT",
                ctypes.c_uint8,
            ),  # 0x9A USB BmAttribute, see http://sdphca.ucsd.edu/lab_equip_manuals/usb_20.pdf, P.266
            (
                "USB_MAX_PWR",
                ctypes.c_uint8,
            ),  # 0x9B USB Max Power, see http://sdphca.ucsd.edu/lab_equip_manuals/usb_20.pdf, P.266, 1 Digit = 2mA
            ("ATT_LB", ctypes.c_uint8),  # 0x9C Attribute Lo byte
            ("ATT_HLB", ctypes.c_uint8),  # 0x9D Attribute Hi Lo byte
            ("ATT_LHB", ctypes.c_uint8),  # 0x9E Attribute Lo Hi byte
            ("ATT_HB", ctypes.c_uint8),  # 0x9F Attribute Hi byte
            ("reserved0", ctypes.c_uint8 * (0xA4 - 0xA0)),  # 0xA0  .. 0xA3 reserved
            ("LUN_PWR_LB", ctypes.c_uint8),  # 0xA4 LUN Power Lo byte
            ("LUN_PWR_HB", ctypes.c_uint8),  # 0xA5 LUN Power Hi byte
            ("reserved1", ctypes.c_uint8 * (0xBF - 0xA6)),  # 0xA6 .. 0xBE reserved
            ("DEV3_ID_STR", ctypes.c_uint8 * (0xC6 - 0xBF)),  # 0xBF .. 0xC5 Card Reader Identifier String
            ("INQ_VEN_STR", ctypes.c_uint8 * (0xCE - 0xC6)),  # 0xC6 .. 0xCD Inquiry Vendor String
            ("INQ_PRD_STR", ctypes.c_uint8 * (0xD3 - 0xCE)),  # 0xCE .. 0xD2 48QFN Inquiry Product String
            ("DYN_NUM_LUN", ctypes.c_uint8),  # 0xD3 Dynamic Number of LUNs
            ("LUN_DEV_MAP", ctypes.c_uint8 * (0xD8 - 0xD4)),  # 0xD4 .. 0xD7 LUN to Device Mapping
            ("reserved2", ctypes.c_uint8 * (0xDB - 0xD8)),  # 0xD8 .. 0xDA reserved
            # HUB CONTROLLER CONFIGURATION
            ("SD_MMC_BUS_TIMING", ctypes.c_uint8 * (0xDE - 0xDB)),  # 0xDB .. 0xDD SD/MMC Bus Timing Control
            ("VID", ctypes.c_uint16),  # 0xDE .. 0xDF Hub Vendor ID
            ("PID", ctypes.c_uint16),  # 0xE0 .. 0xE1 Hub Product ID
            ("DID", ctypes.c_uint16),  # 0xE2 .. 0xE3 Hub Device ID
            ("CFG_DAT_BYTE1", ctypes.c_uint8),  # 0xE4 Configuration Data Byte 1
            ("CFG_DAT_BYTE2", ctypes.c_uint8),  # 0xE5 Configuration Data Byte 2
            ("CFG_DAT_BYTE3", ctypes.c_uint8),  # 0xE6 Configuration Data Byte 3
            ("NR_DEVICE", ctypes.c_uint8),  # 0xE7 Non-Removeable Devices
            ("PORT_DIS_SP", ctypes.c_uint8),  # 0xE8 Port Disable (Self)
            ("PORT_DIS_BP", ctypes.c_uint8),  # 0xE9 Post Disable (Bus)
            ("MAX_PWR_SP", ctypes.c_uint8),  # 0xEA Max Power (Self)
            ("MAX_PWR_BP", ctypes.c_uint8),  # 0xEB Max Power (Bus)
            ("HC_MAX_C_SP", ctypes.c_uint8),  # 0xEC Hub Controller Max Current (Self)
            ("HC_MAX_C_BP", ctypes.c_uint8),  # 0xED Hub Controller Max Current (Bus)
            ("PWR_ON_TIME", ctypes.c_uint8),  # 0xEE Power-on time
            ("BOOST_UP", ctypes.c_uint8),  # 0xEF Boost_Up
            ("BOOST_32", ctypes.c_uint8),  # 0xF0 Boost_3:2
            ("PRT_SWP", ctypes.c_uint8),  # 0xF1 Port Swap
            ("PRTM12", ctypes.c_uint8),  # 0xF2 Port Map 12
            ("PRTM3", ctypes.c_uint8),  # 0xF3 Port Map 3
            # OTHER CONFIGURATION
            ("SD_CLK_LIM", ctypes.c_uint8),  # 0xF4 SD Clock Limit for the Flash Media Controller
            ("reserved3", ctypes.c_uint8),  # 0xF5 reserved
            ("MEDIA_SETTINGS", ctypes.c_uint8),  # 0xF6 SD1 Timeout Configuration
            ("reserved4", ctypes.c_uint8 * (0xFC - 0xF7)),  # 0xF7 .. 0xFB  reserved
            ("NVSTORE_SIG", ctypes.c_uint8 * (0xFF - 0xFC + 1)),  # 0xFC .. 0xFF  Non-Volatile Storage Signature
            # Non Volatile Storage 2 Contents
            ("CLUN0_ID_STR", ctypes.c_uint8 * (0x106 - 0x100 + 1)),  # 0x100 .. 0x106 LUN 0 Identifier String
            ("CLUN1_ID_STR", ctypes.c_uint8 * (0x10D - 0x107 + 1)),  # 0x107 .. 0x10D LUN 1 Identifier String
            ("CLUN2_ID_STR", ctypes.c_uint8 * (0x114 - 0x10E + 1)),  # 0x10E .. 0x114 LUN 2 Identifier String
            ("CLUN3_ID_STR", ctypes.c_uint8 * (0x11B - 0x115 + 1)),  # 0x115 .. 0x11B LUN 3 Identifier String
            ("CLUN4_ID_STR", ctypes.c_uint8 * (0x122 - 0x11C + 1)),  # 0x11C .. 0x122 LUN 4 Identifier String
            ("reserved5", ctypes.c_uint8 * (0x145 - 0x123 + 1)),  # 0x123 .. 0x145 reserved
            ("DYN_NUM_EXT_LUN", ctypes.c_uint8),  # 0x146 Dynamic Number of Extended LUNs
            ("LUN_DEV_MAP2", ctypes.c_uint8 * (0x14B - 0x147 + 1)),  # 0x147 .. 0x14B LUN to Device mapping
            ("reserved6", ctypes.c_uint8 * (0x17B - 0x14C + 1)),  # 0x14C .. 0x17B reserved
            ("NVSTORE_SIG2", ctypes.c_uint8 * (0x17F - 0x17C + 1)),  # 0x17C .. 0x17F Non-Volatile Storage 2 Signature
        ]

        def get_struct(
            reader_VID, reader_PID, reader_vendorString, reader_productString, reader_serial, scsi_mfg, scsi_product
        ):
            """
            Returns a pre-filled EepromStruct.

            Parameters are taken from the Datasheets defaults if nothing else is
            mentioned.
            """
            s = USB2642Eeprom._EepromStruct(
                USB_SER_NUM=string_to_microchip_unicode_uint8_array(reader_serial, 0x19 - 0x00 + 1),
                USB_VID=0x0424,
                USB_PID=0x4041,
                USB_LANG_ID=list_to_uint8_array(
                    [0x04, 0x03, 0x09, 0x04], 0x21 - 0x1E + 1
                ),  # reverse engineered from actual EEPROM. Does NOT match the datasheet.
                USB_MFR_STR=string_to_microchip_unicode_uint8_array(reader_vendorString, 0x5D - 0x22 + 1),
                USB_PRD_STR=string_to_microchip_unicode_uint8_array(reader_productString, 0x99 - 0x5E + 1),
                USB_BM_ATT=0x80,  # Bus Powered, without Remote wakeup
                USB_MAX_PWR=0x30,  # 0x30 * 2mA = 96mA Power Consumption
                ATT_LB=0x50,  # use INQ-strings, SD card is write protected when SW_nWP is high
                ATT_HLB=0x80,
                ATT_LHB=0x00,
                ATT_HB=0x00,
                LUN_PWR_LB=0x00,
                LUN_PWR_HB=0x0A,
                DEV3_ID_STR=string_to_uint8_array("SD/MMC", 0xC5 - 0xBF + 1, encoding="UTF-8", padding=0x00),
                INQ_VEN_STR=string_to_uint8_array(scsi_mfg, 0xCD - 0xC6 + 1, encoding="UTF-8", padding=0x00),
                INQ_PRD_STR=string_to_uint8_array(scsi_product, 0xD2 - 0xCE + 1, encoding="UTF-8", padding=0x00),
                DYN_NUM_LUN=0x01,
                LUN_DEV_MAP=list_to_uint8_array([0xFF, 0x00, 0x00, 0x00], 0xD7 - 0xD4 + 1),
                SD_MMC_BUS_TIMING=list_to_uint8_array([0x59, 0x56, 0x97], 0xDD - 0xDB + 1),
                VID=0x0424,
                PID=0x2640,
                DID=0x08A2,
                CFG_DAT_BYTE1=0x8B,
                CFG_DAT_BYTE2=0x28,
                CFG_DAT_BYTE3=0x00,
                NR_DEVICE=0x02,
                PORT_DIS_SP=0x0C,  # Disable the unused Downstream-Ports of the hub
                PORT_DIS_BP=0x0C,  # Disable the unused Downstream-Ports of the hub
                MAX_PWR_SP=0x01,
                MAX_PWR_BP=0x32,
                HC_MAX_C_SP=0x01,
                HC_MAX_C_BP=0x32,
                PWR_ON_TIME=0x32,
                BOOST_UP=0x00,
                BOOST_32=0x00,
                PORT_SWP=0x00,
                PRTM12=0x00,
                PRTM3=0x00,
                SD_CLK_LIM=0x00,
                MEDIA_SETTINGS=0x00,
                NVSTORE_SIG=string_to_uint8_array("ata2", 4, c_string=False, encoding="UTF-8"),
                # according to datasheet the signature is "ATA2".
                # But reverse engineering the configuration written with the microchip-tool shows
                # that it should be "ata" instead.
                CLUN0_ID_STR=string_to_uint8_array("COMBO", 0x106 - 0x100 + 1, encoding="UTF-8", padding=0x00),
                CLUN1_ID_STR=string_to_uint8_array("COMBO", 0x106 - 0x100 + 1, encoding="UTF-8", padding=0x00),
                CLUN2_ID_STR=string_to_uint8_array("COMBO", 0x106 - 0x100 + 1, encoding="UTF-8", padding=0x00),
                CLUN3_ID_STR=string_to_uint8_array("COMBO", 0x106 - 0x100 + 1, encoding="UTF-8", padding=0x00),
                CLUN4_ID_STR=string_to_uint8_array("COMBO", 0x106 - 0x100 + 1, encoding="UTF-8", padding=0x00),
                DYN_NUM_EXT_LUN=0x00,
                LUN_DEV_MAP2=list_to_uint8_array([0xFF, 0xFF, 0xFF, 0xFF], 0x14B - 0x147 + 1),
                NVSTORE_SIG2=string_to_uint8_array("ecf1", 0x17F - 0x17C + 1, encoding="UTF-8", padding=0x00),
            )
            assert ctypes.sizeof(s) == 384
            return s

    def _read_EEPROM(self, addr=0, len=256):
        """
        Reads len bytes of data starting from addr from the given EEPROM.

        Attributes:
        addr -- Byte address from where to start reading
        len  -- Number of bytes to read (0..256)
        """
        return self.i2c.write_read_to(self.addr, [addr], len)

    def _write_EEPROM(self, addr, data):
        """
        Writes data to the EEPROM starting at addr.

        Attributes:
        addr -- Address to begin write at
        data -- Iterable containing data to write
        """
        offset = 0
        while offset < len(data):
            # determine minimum and maximum address to write in this block
            lower = max((addr + offset) & 0xF0, addr)
            upper = min(((addr + offset) & 0xF0) | 0x0F, addr + len(data))

            lowerOffset = lower - addr
            upperOffset = upper - addr

            self.i2c.write_to(self.addr, [lower] + data[lowerOffset : (upperOffset + 1)])  # noqa: E203
            time.sleep(0.1)
            offset = upperOffset + 1

    def write(self, VID, PID, product_string, vendor_string, serial, scsi_mfg, scsi_product):
        """
        Writes a configuration to the EEPROM.

        Arguments:
        VID -- USB Vendor ID, uint_16
        PID -- USB Product ID, uint_16
        product_string -- Product Name as String
        vendor_string -- Vendor Name as String
        serial -- Serial Number, 12 Hex Digits
        """
        s = USB2642Eeprom._EepromStruct.get_struct(
            reader_VID=VID,
            reader_PID=PID,
            reader_productString=product_string,
            reader_vendorString=vendor_string,
            reader_serial=serial,
            scsi_mfg=scsi_mfg,
            scsi_product=scsi_product,
        )
        buffer = (ctypes.c_uint8 * ctypes.sizeof(s))()
        ctypes.memmove(ctypes.addressof(buffer), ctypes.addressof(s), ctypes.sizeof(s))

        self.i2c.write_config(buffer)


def main():
    parser = argparse.ArgumentParser(
        description="This tool writes and verifies the configuration EEPROM of the usb-sd-mux "
        "with the information given on the command line."
    )
    parser.add_argument("sg", help="The /dev/sg* which is used.")
    parser.add_argument(
        "--productString", help="Sets the product name that will be written.", default="usb-sd-mux_rev1"
    )
    parser.add_argument(
        "--manufacturerString", help="Sets the manufacturerString that will be written.", default="Pengutronix"
    )
    parser.add_argument("--VID", help="Sets the VID that will be written.", default="0x0424")
    parser.add_argument("--ScsiManufacturer", help="Sets the ScsiManufacturer that will be written.", default="PTX")
    parser.add_argument("--ScsiProduct", help="Sets the ScsiProduct that will be written.", default="sdmux")
    parser.add_argument("--PID", help="Sets the USB-PIC that will be written.", default="0x4041")
    parser.add_argument("serial", help="Sets the Serial Number that will be written.")

    args = parser.parse_args()

    c = USB2642Eeprom(args.sg)

    c.write(
        VID=int(args.VID, base=16),
        PID=int(args.PID, base=16),
        product_string=args.productString,
        vendor_string=args.manufacturerString,
        serial=args.serial,
        scsi_mfg=args.ScsiManufacturer,
        scsi_product=args.ScsiProduct,
    )

    print("Write completed")


if __name__ == "__main__":
    main()



import os

import pytest

import usbsdmux.__main__


def test_usage(capsys, mocker):
    "test that the usage output include the command name"
    mocker.patch("sys.argv", ["usbsdmux"])
    with pytest.raises(SystemExit):
        usbsdmux.__main__.main()
    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err.startswith("usage: usbsdmux"), "Invalid help: does not start with 'usage: usbsdmux'"


def test_help_in_readme(capsys, mocker):
    "test that the help output matches the readme"
    mocker.patch("sys.argv", ["usbsdmux", "-h"])
    with pytest.raises(SystemExit):
        usbsdmux.__main__.main()
    captured = capsys.readouterr()
    assert captured.out.startswith("usage: usbsdmux"), "Invalid help: does not start with 'usage: usbsdmux'"
    assert captured.err == "", f"Execution of 'usbsdmux -h' failed: \n{captured.err}"

    readme_path = os.path.join(os.path.dirname(__file__), "../", "README.rst")
    readme_lines = None

    with open(readme_path) as readme:
        for line in readme.readlines():
            line = line.rstrip()
            if line == "   $ usbsdmux -h":
                readme_lines = []
            elif readme_lines is not None:
                if line and not line.startswith("   "):
                    break
                readme_lines.append(line)

    assert readme_lines is not None, "Bash command not found. Did you include '   $ usbsdmux -h'?"
    assert readme_lines, "No output lines found. Did you indent the output correctly?"

    # remove trailing empty lines
    while readme_lines and not readme_lines[-1]:
        readme_lines.pop()

    output_lines = [f"   {line}".rstrip() for line in captured.out.splitlines()]

    assert output_lines == readme_lines, "Output of 'usbsdmux -h' does not match output in README.rst"



import json
import os.path

import pytest

from usbsdmux.sd_regs import CID, SCR, decode_csd

REFS = [
    "02544d53413034471027b7748500bc00",
    "1b534d474638533530d8466363a16700",
    "744a605553442020104182bbc7010600",
    "9f5449303030303000a1114bb5011400",
]


@pytest.mark.parametrize("cid", REFS)
def test_decode(cid):
    ref_name = os.path.join(os.path.dirname(__file__), "reference", f"{cid}.json")

    with open(ref_name) as ref_file:
        ref = json.load(ref_file)

    res = {}
    res["scr"] = SCR(ref["scr"]["raw"]).decode()
    res["cid"] = CID(ref["cid"]["raw"]).decode()
    res["csd"] = decode_csd(ref["csd"]["raw"]).decode()

    # get rid of json differences, like [] -> ()
    res = json.loads(json.dumps(res))

    assert res == ref


@pytest.mark.parametrize("cid", REFS)
def test_to_text(cid):
    ref_name_json = os.path.join(os.path.dirname(__file__), "reference", f"{cid}.json")
    ref_name_text = os.path.join(os.path.dirname(__file__), "reference", f"{cid}.text")

    with open(ref_name_json) as ref_file_json:
        ref_json = json.load(ref_file_json)

    with open(ref_name_text) as ref_file_text:
        ref_text = ref_file_text.read().split("\n")[:-1]

    res = []

    res += decode_csd(ref_json["csd"]["raw"]).get_text_report()
    res += SCR(ref_json["scr"]["raw"]).get_text_report()
    res += CID(ref_json["cid"]["raw"]).get_text_report()

    assert res == ref_text



#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2017 Pengutronix, Chris Fiege <entwicklung@pengutronix.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import ctypes
import fcntl
from time import sleep

from .ctypehelper import (
    list_to_uint8_array,
)

"""
This modules provides an interface to use the auxiliary and configuration
I2C-busses of the Microchip USB2642.
"""


class FrameLengthException(Exception):
    pass


class IoctlFailed(Exception):
    pass


class I2cTransactionFailed(Exception):
    pass


class SDTransactionFailed(Exception):
    pass


class Usb2642:
    """
    This class provides an interface to interact with devices on a Microchip
    USB2642 auxiliary I2C Bus and to write configuration to an EEPROM on the
    configuration I2C Bus.

    To do so it uses vendor specific SCSI-commands on the mass-storage device
    provided by the USB2642.
    Documentation to this behavior can be found in this documents:

    * 'Microchip: I2C_Over_USB_UserGuilde_50002283A.pdf' (Can be found in the
      (Windows-) software example provided on the components webpage)
    * The USB2641 datasheet
    * see http://www.microchip.com/wwwproducts/en/USB2642 for both documents

    Some more interesting links:

    * USB Mass Storage Bulk Transfer Profile Specification:
      http://www.usb.org/developers/docs/devclass_docs/usbmassbulk_10.pdf
    * Linux SG_IO ioctl() control structure:
      http://www.tldp.org/HOWTO/SCSI-Generic-HOWTO/sg_io_hdr_t.html
    * Denton Gentry's blog post about how to use the sg ioctl() from python:
      http://codingrelic.geekhold.com/2012/02/ata-commands-in-python.html
      This article uses it to make ATA-passthrough - beware that we do not use
      ATA-passthrough here.


    This class uses the /dev/sg* -Interface to access the SCSI-device even if no
    media is present.
    Make sure you have rw-rights :)
    """

    def __init__(self, sg):
        """
        Create a new USB2642-Interface wrapper.

        Arguments:
        sg -- The sg-device to use. E.g. "/dev/sg1"
        """
        self.sg = sg

    class _SgioHdrStruct(ctypes.Structure):
        """
        Structure used to access the ioctl() to send arbitrary SCSI-commands.

        Reflects the Kernel-Struct from:
        <scsi/sg.h> sg_io_hdr_t.
        """

        _fields_ = [
            ("interface_id", ctypes.c_int),
            ("dxfer_direction", ctypes.c_int),
            ("cmd_len", ctypes.c_ubyte),
            ("mx_sb_len", ctypes.c_ubyte),
            ("iovec_count", ctypes.c_ushort),
            ("dxfer_len", ctypes.c_uint),
            ("dxferp", ctypes.c_void_p),
            ("cmdp", ctypes.c_void_p),
            ("sbp", ctypes.c_void_p),
            ("timeout", ctypes.c_uint),
            ("flags", ctypes.c_uint),
            ("pack_id", ctypes.c_int),
            ("usr_ptr", ctypes.c_void_p),
            ("status", ctypes.c_ubyte),
            ("masked_status", ctypes.c_ubyte),
            ("msg_status", ctypes.c_ubyte),
            ("sb_len_wr", ctypes.c_ubyte),
            ("host_status", ctypes.c_ushort),
            ("driver_status", ctypes.c_ushort),
            ("resid", ctypes.c_int),
            ("duration", ctypes.c_uint),
            ("info", ctypes.c_uint),
        ]

    # sg_io_hdr_t contains 9 ints, 3 short ints, 6 chars and 4 pointers. So its
    # size is 9 * 4 + 3 * 2 + 6 * 1 + 4 * 4 = 64 on 32 bit architectures. On 64
    # bit architectures there are two holes in the struct:
    # - 4 bytes before *usr_ptr to make the pointer aligned
    # - 4 bytes at the end to make the size a multiple of 8.
    # So the size there is: 9 * 4 + 3 * 2 + 6 * 1 + 4 * 8 + 2 * 4 = 88.
    if ctypes.sizeof(ctypes.c_void_p) == 4:
        assert ctypes.sizeof(_SgioHdrStruct) == 64
    else:
        assert ctypes.sizeof(ctypes.c_void_p) == 8
        assert ctypes.sizeof(_SgioHdrStruct) == 88

    """IOCTL for SG_IO"""
    _SG_IO = 0x2285  # <scsi/sg.h>

    """SgioHdr dxfer direction constant: No direction"""
    _SG_DXFER_NONE = -1

    """SgioHdr dxfer direction constant: Host to device"""
    _SG_DXFER_TO_DEV = -2

    """SgioHdr dxfer direction constant: Device to Host"""
    _SG_DXFER_FROM_DEV = -3

    """
    This Opcode represents a vendor specific SCSI command.
    According to: 'Microchip: I2C_Over_USB_UserGuilde_50002283A.pdf' P.20
    """
    _USB2642SCSIOPCODE = 0xCF

    """
    This Vendor Action marks an I2C Write Action
    According to: 'Microchip: I2C_Over_USB_UserGuilde_50002283A.pdf' P.20
    """
    _USB2642I2CWRITESTREAM = 0x23

    """
    This Vendor Action marks an I2C Write-Read Action
    According to: 'Microchip: I2C_Over_USB_UserGuilde_50002283A.pdf' P.20
    """
    _USB2642I2CWRITEREADSTREAM = 0x22

    class _USB2642I2cWriteStruct(ctypes.Structure):
        """I2C-Write Data Structure for up to 512 Bytes of Data

        According to: 'Microchip: I2C_Over_USB_UserGuilde_50002283A.pdf' P.20
        """

        _fields_ = [
            ("ScsiVendorCommand", ctypes.c_uint8),
            ("ScsiVendorActionWriteI2C", ctypes.c_uint8),
            ("I2cSlaveAddress", ctypes.c_uint8),
            ("I2cUnused", ctypes.c_uint8),
            ("I2cDataPhaseLenHigh", ctypes.c_uint8),
            ("I2cDataPhaseLenLow", ctypes.c_uint8),
            ("I2cCommandPhaseLen", ctypes.c_uint8),
            ("I2cCommandPayload", ctypes.c_uint8 * 9),
        ]

    assert ctypes.sizeof(_USB2642I2cWriteStruct) == 16

    class _USB2642I2cReadStruct(ctypes.Structure):
        """
        I2C-Read Data Structure for up to 512 Bytes of Data.

        According to: 'Microchip: I2C_Over_USB_UserGuilde_50002283A.pdf' P.20
        """

        _fields_ = [
            ("ScsiVendorCommand", ctypes.c_uint8),
            ("ScsiVendorActionWriteReadI2C", ctypes.c_uint8),
            ("I2cWriteSlaveAddress", ctypes.c_uint8),
            ("I2cReadSlaveAddress", ctypes.c_uint8),
            ("I2cReadPhaseLenHigh", ctypes.c_uint8),
            ("I2cReadPhaseLenLow", ctypes.c_uint8),
            ("I2cWritePhaseLen", ctypes.c_uint8),
            ("I2cWritePayload", ctypes.c_uint8 * 9),
        ]

    assert ctypes.sizeof(_USB2642I2cReadStruct) == 16

    def _get_SCSI_cmd_I2C_write(self, slaveAddr, data):
        """
        Create an I2cWrite Command Structure to write up to 512 bytes to device
        slaveAddr.

        According to: 'Microchip: I2C_Over_USB_UserGuilde_50002283A.pdf' P.20
        """
        MAXLEN = 512
        count = min(len(data), MAXLEN)
        dataArray = (ctypes.c_uint8 * MAXLEN)()
        dataArray[:count] = data[:count]

        slaveWriteAddr = (slaveAddr * 2) & 0xFF

        cmd = self._USB2642I2cWriteStruct(
            ScsiVendorCommand=self._USB2642SCSIOPCODE,
            ScsiVendorActionWriteI2C=self._USB2642I2CWRITESTREAM,
            I2cSlaveAddress=slaveWriteAddr,
            I2cUnused=0x00,
            I2cDataPhaseLenHigh=(count >> 8) & 0xFF,
            I2cDataPhaseLenLow=count & 0xFF,
            I2cCommandPhaseLen=0x00,
            I2cCommandPayload=(ctypes.c_uint8 * 9)(),
        )

        return cmd, dataArray

    def _get_SCSI_cmd_I2C_write_read(self, slaveAddr, writeData, readLength):
        """
        Create an I2cWriteRead Command Structure to write up to 9 bytes to device
        slaveAddr and then read back up to 512 bytes of data.

        According to: 'Microchip: I2C_Over_USB_UserGuilde_50002283A.pdf' P.20
        """
        MAXLEN = 512
        readCount = min(readLength, MAXLEN)
        readDataArray = (ctypes.c_uint8 * MAXLEN)()

        MAXLEN = 9
        writeCount = min(len(writeData), MAXLEN)
        writeDataArray = (ctypes.c_uint8 * MAXLEN)()
        writeDataArray[:writeCount] = writeData[:writeCount]

        slaveWriteAddr = (slaveAddr * 2) & 0xFF
        slaveReadAddr = slaveWriteAddr + 1

        cmd = self._USB2642I2cReadStruct(
            ScsiVendorCommand=self._USB2642SCSIOPCODE,
            ScsiVendorActionWriteReadI2C=self._USB2642I2CWRITEREADSTREAM,
            I2cWriteSlaveAddress=slaveWriteAddr,
            I2cReadSlaveAddress=slaveReadAddr,
            I2cReadPhaseLenHigh=(readCount >> 8) & 0xFF,
            I2cReadPhaseLenLow=readCount & 0xFF,
            I2cWritePhaseLen=writeCount,
            I2cWritePayload=writeDataArray,
        )

        return cmd, readDataArray

    def _get_SGIO(self, command, Sg_Dxfer, databuffer):
        """Fill the SG_IO ioctl() -structure with sane defaults for the given
        command.
        The command will create a 64-byte sense-buffer for returned status.

        Arguments:
        command -- SCSI-Command to use. ctypes.c_uint8
        Sg_Dxfer -- _SG_DXFER_* to set the direction of the SCSI transfer.
        databuffer -- 512 bytes buffer of the block to read or write.
        """
        sense = ctypes.c_buffer(64)

        sgio = self._SgioHdrStruct(
            # "S" for SCSI
            interface_id=ord("S"),
            # SG_DXFER_*
            dxfer_direction=Sg_Dxfer,
            # length of whatever we put into cmd
            cmd_len=ctypes.sizeof(command),
            # length of sense buffer
            mx_sb_len=ctypes.sizeof(sense),
            iovec_count=0,
            # data transfer length
            dxfer_len=ctypes.sizeof(databuffer),
            # pointer to data transfer buffer
            dxferp=ctypes.cast(databuffer, ctypes.c_void_p),
            # command to perform
            cmdp=ctypes.cast(ctypes.addressof(command), ctypes.c_void_p),
            # sense buffer memory
            sbp=ctypes.cast(sense, ctypes.c_void_p),
            # a timeout for this command in ms
            timeout=3000,
            # SG_FLAG_*, normally 0
            flags=0,
            # unused
            pack_id=0,
            # unused
            usr_ptr=None,
            # output: SCSI-status
            status=0,
            # output: shifted, maskes SCSI-stauts
            masked_status=0,
            # output: optional: message level data
            msg_status=0,
            # output: byte actually written to sbp
            sb_len_wr=0,
            # output: errors from host adapter
            host_status=0,
            # output: errors from software driver
            driver_status=0,
            # output: result_len: actually transferred data
            resid=0,
            # output: time for the command in ms
            duration=0,
            # output: auxiliary information (?)
            info=0,
        )

        return sgio, sense

    def _call_IOCTL(self, command, sg_dxfer, databuffer):
        """
        Call the ioctl()

        This function will create the struct to call the ioctl() and handle return
        codes.

        Arguments:
        command -- SCSI Command payload to send. 16-Byte buffer containing the SCSI
                   command parameters.
        sg_dxfer -- _SG_DXFER_*: Direction of the SCSI transfer
        databuffer -- 512 byte long buffer to be written or read
        """
        sgio, sense = self._get_SGIO(command, sg_dxfer, databuffer)
        #    print("SGIO:")
        #    print(self.to_pretty_hex(sgio))

        with open(self.sg, "r+b", buffering=0) as fh:
            rc = fcntl.ioctl(fh, self._SG_IO, sgio)
            if rc != 0:
                raise IoctlFailed(f"SG_IO ioctl() failed with non-zero exit-code {rc}")
        return databuffer, sense, sgio

    def write_config(self, data):
        """
        Writes the eeprom contents from data into the config EEPROM on the auxiliary
        I2C bus.

        This is done using reverse-engineered commands send by the Microchip
        Windows-Demo-Tool.

        Arguments:
        data -- EEPROM blob to write as ctype.buffer. Length 384 Bytes as described
                in the USB2642 Datasheet.
        """

        # SCSI Command was found on the USB-Bus.
        # Since most of the bytes are unknown this is used as plain magic.
        scsiCommand = list_to_uint8_array(
            [0xCF, 0x54, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00], 16
        )

        # Data in the captured USB-transfer was suffixed with some random data.
        # Experiments showed that 0x00 works fine too.
        # Since the buffer is zero-ed when initialized the suffix could be removed.
        data_suffix = list_to_uint8_array([0x00], 127)

        # Copying prefix, data and suffix to the SCSI command data-section
        payload = (ctypes.c_uint8 * 512)()
        payload[: ctypes.sizeof(data)] = data
        payload[ctypes.sizeof(data) : ctypes.sizeof(data) + ctypes.sizeof(data_suffix)] = data_suffix  # noqa: E203

        # Perform the actual SCSI transfer
        self._call_IOCTL(scsiCommand, self._SG_DXFER_TO_DEV, payload)

    def write_read_to(self, i2cAddr, writeData, readLength):
        """
        Tries to write data to an I2C-Device and afterwards read data from that
        device.

        This function will perform am I2C-Transaction like the following:

        * I2C-Start
        * I2C-Slave address with R/W = W (0)
        * writeData[0]
        * writeData[1]
        * ...
        * I2C-Repeated Start
        * I2C-Slave address with R/W = R (1)
        * readData[0]
        * readData[1]
        * ...
        * I2C-Stop

        This transaction can (for example) be used to set the address-pointer inside
        an EEPROM and read data from it.

        Arguments:
        i2cAddr -- 7-Bit I2C Slave address (as used by Linux). Will be shifted 1 Bit
                   to the left before adding the R/W-bit.
        writeData -- iterable of bytes to write in the first phase
        readLengh -- number of bytes (0..512) to read in the second phase
        """
        scsiCommand, data = self._get_SCSI_cmd_I2C_write_read(i2cAddr, writeData, readLength)
        # TODO: Add error handling if length of read or write do not match
        #       requirements

        #    print("I2C-Command:")
        #    print(self.to_pretty_hex(scsiCommand))
        #    print("I2C-Payload:")
        #    print(self.to_pretty_hex(data))
        data, sense, sgio = self._call_IOCTL(scsiCommand, self._SG_DXFER_FROM_DEV, data)

        if sgio.status != 0:
            raise I2cTransactionFailed(
                f"SCSI-Transaction ended with status {sgio.status}. I2C-Transaction has probably failed."
            )

        return list(data[:readLength])

    def write_to(self, i2cAddr, data):
        """
        Tries to write data to an I2C-Device.

        This function will perform am I2C-Transaction like the following:

        * I2C-Start
        * I2C-Slave address with R/W = W (0)
        * data[0]
        * data[1]
        * ...
        * I2C-Stop

        Transactions like this can (for example) be used if configuration registers
        on a device have to be written.

        Arguments:
        i2cAddr -- 7-Bit I2C Slave address (as used by Linux). Will be shifted 1 Bit
                   to the left before adding the R/W-bit.
        data -- iterateable of bytes to write."""
        scsiCommand, data = self._get_SCSI_cmd_I2C_write(i2cAddr, data)
        # TODO: Add length checks

        #    print("I2C-Command:")
        #    print(self.to_pretty_hex(scsiCommand))
        #    print("I2C-Payload:")
        #    print(self.to_pretty_hex(data))
        data, sense, sgio = self._call_IOCTL(scsiCommand, self._SG_DXFER_TO_DEV, data)

        if sgio.status != 0:
            raise I2cTransactionFailed(
                f"SCSI-Transaction ended with status {sgio.status}. I2C-Transaction has probably failed."
            )

    def _read_register(self, reg, size, retries=5):
        scsiCommand = list_to_uint8_array(
            [0xCF, reg, 0x00, 0x00, size, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00], 16
        )

        while True:
            databuffer = ctypes.c_buffer(size)
            _, _, sgio = self._call_IOCTL(scsiCommand, self._SG_DXFER_FROM_DEV, databuffer)

            if retries and sgio.status == 2:
                sleep(0.5)
                retries -= 1
                continue

            if sgio.status != 0:
                raise SDTransactionFailed(
                    f"SCSI Transaction ended with status {sgio.status}. SD Transaction has probably failed."
                )

            break

        return databuffer.raw

    def read_cid(self):
        return self._read_register(0x18, 16)

    def read_csd(self):
        return self._read_register(0x1A, 16)

    def read_scr(self):
        return self._read_register(0x1B, 8)
